==> apache-inlong-1.8.0-src.tar.gz: apache-inlong-1.8.0/inlong-sort/sort-flink/cdc-base/src/main/java/org/apache/inlong/sort/cdc/base/util/RecordUtils.java <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.inlong.sort.cdc.base.util;

import io.debezium.connector.AbstractSourceInfo;
import io.debezium.data.Envelope;
import io.debezium.relational.Column;
import io.debezium.relational.TableId;
import org.apache.flink.table.types.logical.BigIntType;
import org.apache.flink.table.types.logical.BinaryType;
import org.apache.flink.table.types.logical.BooleanType;
import org.apache.flink.table.types.logical.DecimalType;
import org.apache.flink.table.types.logical.DoubleType;
import org.apache.flink.table.types.logical.FloatType;
import org.apache.flink.table.types.logical.IntType;
import org.apache.flink.table.types.logical.LogicalType;
import org.apache.flink.table.types.logical.SmallIntType;
import org.apache.flink.table.types.logical.TimestampType;
import org.apache.flink.table.types.logical.TinyIntType;
import org.apache.flink.table.types.logical.VarCharType;
import org.apache.kafka.connect.data.Schema;
import org.apache.kafka.connect.data.Struct;
import org.apache.kafka.connect.source.SourceRecord;

import java.util.Arrays;
import java.util.List;

import static org.apache.inlong.sort.cdc.base.relational.JdbcSourceEventDispatcher.HISTORY_RECORD_FIELD;

/**
 * Utility class to deal record.
 */
public class RecordUtils {

    private static final List<String> NUMBER_TYPE = Arrays.asList("NUMBER");
    private static final List<String> FLOAT_TYPE = Arrays.asList("FLOAT", "BINARY_FLOAT");
    private static final List<String> DOUBLE_TYPE = Arrays.asList("DOUBLE PRECISION", "BINARY_DOUBLE");
    private static final List<String> TIMESTAMP_TYPE = Arrays
            .asList("DATE", "TIMESTAMP", "WITH LOCAL TIME ZONE", "TIMESTAMP WITH TIME ZONE");
    private static final List<String> VARCHAR_TYPE = Arrays
            .asList("CHAR", "NCHAR", "NVARCHAR2", "NVCHAER", "VARCHAR", "VARCHAR2", "CLOB", "NCLOB", "XMLType");
    private static final List<String> BINARY_TYPE = Arrays.asList("BLOB", "ROWID");
    private static final List<String> BIGINT_TYPE = Arrays.asList("INTERVAL DAY TO SECOND", "INTERVAL YEAR TO MONTH");
    public static final String MYSQL_SCHEMA_CHANGE_EVENT_KEY_NAME = "io.debezium.connector.mysql.SchemaChangeKey";
    public static final String ORACLE_SCHEMA_CHANGE_EVENT_KEY_NAME = "io.debezium.connector.oracle.SchemaChangeKey";
    public static final String CONNECTOR = "connector";
    public static final String MYSQL_CONNECTOR = "mysql";

    private RecordUtils() {

    }

    /**
     * Get table id from source record
     * @param dataRecord
     * @return
     */
    public static TableId getTableId(SourceRecord dataRecord) {
        Struct value = (Struct) dataRecord.value();
        Struct source = value.getStruct(Envelope.FieldName.SOURCE);
        String dbName = source.getString(AbstractSourceInfo.DATABASE_NAME_KEY);
        String schemaName = source.getString(AbstractSourceInfo.SCHEMA_NAME_KEY);
        String tableName = source.getString(AbstractSourceInfo.TABLE_NAME_KEY);
        return new TableId(dbName, schemaName, tableName);
    }

    /**
     * According column's type, get it's logicalType
     * Refer: https://ververica.github.io/flink-cdc-connectors/master/content/connectors/oracle-cdc.html?from_wecom=1
     * @param column
     * @param struct
     * @return
     */
    public static LogicalType convertLogicType(Column column, Struct struct) {
        String typeName = column.typeName();
        // Oracle Number Type default precision is 38
        Integer p = column.length() == 0 ? 38 : column.length();
        Integer s = column.scale().isPresent() ? column.scale().orElse(0) : struct.getInt32("scale");
        if (NUMBER_TYPE.contains(typeName)) {
            if (p == 1) {
                return new BooleanType();
            }
            if (s <= 0 && p - s < 3) {
                return new TinyIntType();
            }
            if (s <= 0 && p - s < 5) {
                return new SmallIntType();
            }
            if (s <= 0 && p - s < 10) {
                return new IntType();
            }
            if (s <= 0 && p - s < 19) {
                return new BigIntType();
            }
            if (s <= 0 && p - s >= 10 && p - s <= 38) {
                return new DecimalType(p - s, 0);
            }
            if (s > 0) {
                return new DecimalType(p, s);
            }
            if (s <= 0 && p - s > 38) {
                return new VarCharType(Integer.MAX_VALUE);
            }
        }
        if (FLOAT_TYPE.contains(typeName)) {
            return new FloatType();
        }
        if (DOUBLE_TYPE.contains(typeName)) {
            return new DoubleType();
        }
        if (TIMESTAMP_TYPE.contains(typeName)) {
            return new TimestampType(p);
        }
        if (VARCHAR_TYPE.contains(typeName)) {
            return new VarCharType(Integer.MAX_VALUE);
        }
        if (BINARY_TYPE.contains(typeName)) {
            return new BinaryType();
        }
        if (BIGINT_TYPE.contains(typeName)) {
            return new BigIntType();
        }
        return null;
    }

    /**
     * Whether the source Record is a schema change event.
     * @param sourceRecord
     * @return ture if the source Record is a schema change event.
     */
    public static boolean isSchemaChangeEvent(SourceRecord sourceRecord) {
        Schema keySchema = sourceRecord.keySchema();
        return keySchema != null && (MYSQL_SCHEMA_CHANGE_EVENT_KEY_NAME.equalsIgnoreCase(keySchema.name())
                || ORACLE_SCHEMA_CHANGE_EVENT_KEY_NAME.equalsIgnoreCase(keySchema.name()));
    }

    /**
     * Whether the source belong Mysql Connector
     * @param source
     * @return true if the source belong Mysql Connector
     */
    public static boolean isMysqlConnector(Struct source) {
        String connector = source.getString(CONNECTOR);
        return MYSQL_CONNECTOR.equalsIgnoreCase(connector);
    }

    public static boolean isDdlRecord(Struct value) {
        return value.schema().field(HISTORY_RECORD_FIELD) != null;
    }

}

==> apache-inlong-2.3.0-src.tar.gz: apache-inlong-2.3.0/inlong-sort/sort-flink/cdc-base/src/main/java/org/apache/inlong/sort/cdc/base/util/RecordUtils.java <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.inlong.sort.cdc.base.util;

import io.debezium.connector.AbstractSourceInfo;
import io.debezium.data.Envelope;
import io.debezium.relational.Column;
import io.debezium.relational.TableId;
import org.apache.flink.table.types.logical.BigIntType;
import org.apache.flink.table.types.logical.BinaryType;
import org.apache.flink.table.types.logical.BooleanType;
import org.apache.flink.table.types.logical.DecimalType;
import org.apache.flink.table.types.logical.DoubleType;
import org.apache.flink.table.types.logical.FloatType;
import org.apache.flink.table.types.logical.IntType;
import org.apache.flink.table.types.logical.LogicalType;
import org.apache.flink.table.types.logical.SmallIntType;
import org.apache.flink.table.types.logical.TimestampType;
import org.apache.flink.table.types.logical.TinyIntType;
import org.apache.flink.table.types.logical.VarCharType;
import org.apache.kafka.connect.data.Schema;
import org.apache.kafka.connect.data.Struct;
import org.apache.kafka.connect.source.SourceRecord;

import java.util.Arrays;
import java.util.List;

import static org.apache.inlong.sort.cdc.base.relational.JdbcSourceEventDispatcher.HISTORY_RECORD_FIELD;

/**
 * Utility class to deal record.
 */
public class RecordUtils {

    private static final List<String> NUMBER_TYPE = Arrays.asList("NUMBER");
    private static final List<String> FLOAT_TYPE = Arrays.asList("FLOAT", "BINARY_FLOAT");
    private static final List<String> DOUBLE_TYPE = Arrays.asList("DOUBLE PRECISION", "BINARY_DOUBLE");
    private static final List<String> TIMESTAMP_TYPE = Arrays
            .asList("DATE", "TIMESTAMP", "WITH LOCAL TIME ZONE", "TIMESTAMP WITH TIME ZONE");
    private static final List<String> VARCHAR_TYPE = Arrays
            .asList("CHAR", "NCHAR", "NVARCHAR2", "NVCHAER", "VARCHAR", "VARCHAR2", "CLOB", "NCLOB", "XMLType");
    private static final List<String> BINARY_TYPE = Arrays.asList("BLOB", "ROWID");
    private static final List<String> BIGINT_TYPE = Arrays.asList("INTERVAL DAY TO SECOND", "INTERVAL YEAR TO MONTH");
    public static final String SCHEMA_CHANGE_EVENT_KEY_NAME = "io.debezium.connector.*.SchemaChangeKey";
    public static final String CONNECTOR = "connector";
    public static final String MYSQL_CONNECTOR = "mysql";
    public static final String SCHEMA_HEARTBEAT_EVENT_KEY_NAME = "io.debezium.connector.common.Heartbeat";

    private RecordUtils() {

    }

    /**
     * Get table id from source record
     * @param dataRecord
     * @return
     */
    public static TableId getTableId(SourceRecord dataRecord) {
        Struct value = (Struct) dataRecord.value();
        Struct source = value.getStruct(Envelope.FieldName.SOURCE);
        String dbName = source.getString(AbstractSourceInfo.DATABASE_NAME_KEY);
        String schemaName = source.getString(AbstractSourceInfo.SCHEMA_NAME_KEY);
        String tableName = source.getString(AbstractSourceInfo.TABLE_NAME_KEY);
        return new TableId(dbName, schemaName, tableName);
    }

    /**
     * According column's type, get it's logicalType
     * Refer: https://ververica.github.io/flink-cdc-connectors/master/content/connectors/oracle-cdc.html?from_wecom=1
     * @param column
     * @param struct
     * @return
     */
    public static LogicalType convertLogicType(Column column, Struct struct) {
        String typeName = column.typeName();
        // Oracle Number Type default precision is 38
        Integer p = column.length() == 0 ? 38 : column.length();
        Integer s = column.scale().isPresent() ? column.scale().orElse(0) : struct.getInt32("scale");
        if (NUMBER_TYPE.contains(typeName)) {
            if (p == 1) {
                return new BooleanType();
            }
            if (s <= 0 && p - s < 3) {
                return new TinyIntType();
            }
            if (s <= 0 && p - s < 5) {
                return new SmallIntType();
            }
            if (s <= 0 && p - s < 10) {
                return new IntType();
            }
            if (s <= 0 && p - s < 19) {
                return new BigIntType();
            }
            if (s <= 0 && p - s >= 10 && p - s <= 38) {
                return new DecimalType(p - s, 0);
            }
            if (s > 0) {
                return new DecimalType(p, s);
            }
            if (s <= 0 && p - s > 38) {
                return new VarCharType(Integer.MAX_VALUE);
            }
        }
        if (FLOAT_TYPE.contains(typeName)) {
            return new FloatType();
        }
        if (DOUBLE_TYPE.contains(typeName)) {
            return new DoubleType();
        }
        if (TIMESTAMP_TYPE.contains(typeName)) {
            return new TimestampType(p);
        }
        if (VARCHAR_TYPE.contains(typeName)) {
            return new VarCharType(Integer.MAX_VALUE);
        }
        if (BINARY_TYPE.contains(typeName)) {
            return new BinaryType();
        }
        if (BIGINT_TYPE.contains(typeName)) {
            return new BigIntType();
        }
        return null;
    }

    /**
     * Whether the source Record is a schema change event.
     * @param sourceRecord
     * @return ture if the source Record is a schema change event.
     */
    public static boolean isSchemaChangeEvent(SourceRecord sourceRecord) {
        Schema keySchema = sourceRecord.keySchema();
        return keySchema != null && (keySchema.name().matches(SCHEMA_CHANGE_EVENT_KEY_NAME));
    }

    public static boolean isDdlRecord(Struct value) {
        return value.schema().field(HISTORY_RECORD_FIELD) != null;
    }

    public static boolean isHeartbeatEvent(SourceRecord record) {
        Schema valueSchema = record.valueSchema();
        return valueSchema != null
                && SCHEMA_HEARTBEAT_EVENT_KEY_NAME.equalsIgnoreCase(valueSchema.name());
    }

    public static boolean shouldUseCatalogBeforeSchema(TableId tableId) {
        // if catalog is not defined but the schema is defined return this flag as false
        // otherwise return true
        return !(tableId.catalog() == null && tableId.schema() != null);
    }

}
